/**
 *  MEDOR: Middleware Enabling Distributed Object Requests
 *
 * Copyright (C) 2001-2005 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, S. Chassande-Barrioz, A. Lefebvre
 */
package org.objectweb.medor.optim.jorm.rdb;

import junit.framework.Test;
import junit.framework.TestSuite;
import junit.textui.TestRunner;
import org.objectweb.jorm.api.PException;
import org.objectweb.jorm.api.PMapper;
import org.objectweb.jorm.mapper.rdb.lib.ConnectionSpecJDBC;
import org.objectweb.jorm.mapper.rdb.lib.MapperJDBC;
import org.objectweb.jorm.mapper.rdb.adapter.api.RdbAdapter;
import org.objectweb.jorm.mapper.rdb.adapter.api.RdbAdapterException;
import org.objectweb.jorm.mapper.rdb.metainfo.RdbClassMapping;
import org.objectweb.jorm.mapper.rdb.adapter.RdbAdapterFactory;
import org.objectweb.jorm.type.api.PTypeSpace;
import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.expression.lib.And;
import org.objectweb.medor.expression.lib.BasicOperand;
import org.objectweb.medor.expression.lib.BasicParameterOperand;
import org.objectweb.medor.expression.lib.Equal;
import org.objectweb.medor.expression.lib.Or;
import org.objectweb.medor.filter.lib.BasicFieldOperand;
import org.objectweb.medor.filter.lib.MemberOf;
import org.objectweb.medor.lib.Log;
import org.objectweb.medor.optim.api.QueryRewriter;
import org.objectweb.medor.optim.jorm.Jorm2Rdb;
import org.objectweb.medor.optim.jorm.JormFilteredInheritance;
import org.objectweb.medor.optim.jorm.TestJormQueryRewriter;
import org.objectweb.medor.optim.lib.BasicQueryRewriter;
import org.objectweb.medor.optim.lib.FlattenQueryTreeRule;
import org.objectweb.medor.query.api.OrderField;
import org.objectweb.medor.query.api.QueryLeaf;
import org.objectweb.medor.query.lib.BasicOrderField;
import org.objectweb.medor.query.rdb.api.QualifiedTable;
import org.objectweb.medor.query.rdb.api.RdbExpField;
import org.objectweb.medor.query.rdb.lib.BasicQualifiedTable;
import org.objectweb.medor.query.rdb.lib.BasicRdbExpQueryLeaf;

import java.util.ArrayList;

/**
 * @author S.Chassande-Barrioz
 */
public class TestRdbJormQueryRewriter extends TestJormQueryRewriter {

	RdbAdapter adapter;

	/**
	 * main method to launch the tests manually
	 */
	public static void main(String[] args) {
		TestRunner.run(suite());
	}

	/**
	 * This method creates a TestSuite object with the current tests
	 */
	public static Test suite() {
		return new TestSuite(TestRdbJormQueryRewriter.class);
	}

	public TestRdbJormQueryRewriter() {
		this("TestRdbJormQueryRewriter");
	}

	public TestRdbJormQueryRewriter(String testName) {
		super(testName, Log.MEDOR_PREFIX + ".optim.jorm.rdb.queryrewriter");
		try {
			adapter = RdbAdapterFactory.getTypeConverter((String) null);
		} catch (RdbAdapterException ex) {
			fail(ex.getMessage());
		}
	}

	public PMapper newMappper() {
		PMapper m = null;
		try {
            m = new MapperJDBC();
			ConnectionSpecJDBC cs = new ConnectionSpecJDBC("jdbc:product:basename", "java.lang.String");
			m.setConnectionFactory(cs);
		} catch (PException e) {
			e.printStackTrace();
		}
		return m;
	}

	public QueryRewriter createQueryRewriter() {
		ArrayList rules = new ArrayList(2);
		rules.add(new JormFilteredInheritance());
		rules.add(new FlattenQueryTreeRule());
		rules.add(new Jorm2Rdb());
		return new BasicQueryRewriter(rules);
	}

	public QueryLeaf getQueryLeafA(Example e) throws MedorException {
		RdbClassMapping cm = (RdbClassMapping)
				e.clazz.getClassProject(getProjectName()).getMapping(getMapperName()).getClassMapping();
		String name = e.clazz.getFQName();
		String prefix = name + ".";

		QualifiedTable qt = new BasicQualifiedTable(cm.getRdbTable().getName(), "T_REF_SING_A_0");
		BasicRdbExpQueryLeaf leaf = new BasicRdbExpQueryLeaf(
				null, new QualifiedTable[]{qt}, "");
		leaf.addRdbField(prefix + "id", PTypeSpace.LONG, "id_col", qt);
		RdbExpField f1 = leaf.addRdbField(prefix + "field1", PTypeSpace.STRING, "FIELD_1", qt);
		leaf.removeRdbField(f1);
		leaf.setQueryFilter(new Equal(
				new BasicFieldOperand(f1),
				new BasicParameterOperand(PTypeSpace.STRING, "?1")));
		leaf.setRdbAdapter(adapter);
		return leaf;
	}

	public QueryLeaf getQueryLeafLongLong(Example e) throws MedorException {
		RdbClassMapping cm = (RdbClassMapping)
				e.clazz.getClassProject(getProjectName()).getMapping(getMapperName()).getClassMapping();
		String name = e.clazz.getFQName();
		String prefix = name + ".";

		QualifiedTable qt = new BasicQualifiedTable(cm.getRdbTable().getName(), "T_B_0");
		BasicRdbExpQueryLeaf leaf = new BasicRdbExpQueryLeaf(
				null, new QualifiedTable[]{qt}, "");
		RdbExpField f1 = leaf.addRdbField(prefix + "c1", PTypeSpace.LONG, "C1", qt);
		RdbExpField f2 = leaf.addRdbField(prefix + "c2", PTypeSpace.LONG, "C2", qt);
		leaf.setQueryFilter(new And(
				new Equal(
					new BasicFieldOperand(f1),
					new BasicOperand(-1)),
				new Equal(
					new BasicFieldOperand(f2),
					new BasicOperand(-1))
				));
		leaf.setRdbAdapter(adapter);
		return leaf;
	}

	public QueryLeaf getQueryLeafAgg(Example e) throws MedorException {
		RdbClassMapping cm = (RdbClassMapping)
				e.clazz.getClassProject(getProjectName()).getMapping(getMapperName()).getClassMapping();
		String name = e.clazz.getFQName();
		String prefix = name + ".";

		QualifiedTable qt = new BasicQualifiedTable(cm.getRdbTable().getName(), "T_Agg_0");
		BasicRdbExpQueryLeaf leaf = new BasicRdbExpQueryLeaf(
				null, new QualifiedTable[]{qt}, "");
		leaf.addRdbField(prefix + "c2", PTypeSpace.LONG, "C2", qt);
		leaf.addRdbField(prefix + "c1", PTypeSpace.INT, "C1", qt);
		leaf.addRdbField(prefix + "y", PTypeSpace.FLOAT, "Y", qt);
		leaf.setRdbAdapter(adapter);
		return leaf;
	}

	public QueryLeaf getQueryLeafAggParam(Example e) throws MedorException {
		RdbClassMapping cm = (RdbClassMapping)
				e.clazz.getClassProject(getProjectName()).getMapping(getMapperName()).getClassMapping();
		String name = e.clazz.getFQName();
		String prefix = name + ".";

		QualifiedTable qt = new BasicQualifiedTable(cm.getRdbTable().getName(), "T_REF_SING_A_0");
		BasicRdbExpQueryLeaf leaf = new BasicRdbExpQueryLeaf(
				null, new QualifiedTable[]{qt}, "");
		leaf.addRdbField(prefix + "id", PTypeSpace.LONG, "id_col", qt);
		leaf.setRdbAdapter(adapter);
		return leaf;
	}

	public QueryLeaf getQueryLeafAggDirect(Example e) throws MedorException {
		RdbClassMapping cm = (RdbClassMapping)
				e.clazz.getClassProject(getProjectName()).getMapping(getMapperName()).getClassMapping();
		String name = e.clazz.getFQName();
		String prefix = name + ".";

		QualifiedTable qt = new BasicQualifiedTable(cm.getRdbTable().getName(), "T_Agg_0");
		BasicRdbExpQueryLeaf leaf = new BasicRdbExpQueryLeaf(
				null, new QualifiedTable[]{qt}, "");
		leaf.addRdbField(prefix + "c1", PTypeSpace.INT, "C1", qt);
		leaf.addRdbField(prefix + "y", PTypeSpace.FLOAT, "Y", qt);
		leaf.setRdbAdapter(adapter);
		return leaf;
	}

	public QueryLeaf getQueryLeafAggWithWhere(Example e) throws MedorException {
		RdbClassMapping cm = (RdbClassMapping)
				e.clazz.getClassProject(getProjectName()).getMapping(getMapperName()).getClassMapping();
		String name = e.clazz.getFQName();
		String prefix = name + ".";

		QualifiedTable qt = new BasicQualifiedTable(cm.getRdbTable().getName(), "T_Agg_0");
		BasicRdbExpQueryLeaf leaf = new BasicRdbExpQueryLeaf(
				null, new QualifiedTable[]{qt}, "");
		RdbExpField c2 = leaf.addRdbField(prefix + "c2", PTypeSpace.LONG, "C2", qt);
		leaf.removeRdbField(c2);
		leaf.addRdbField(prefix + "c1", PTypeSpace.INT, "C1", qt);
		leaf.addRdbField(prefix + "y", PTypeSpace.FLOAT, "Y", qt);
		leaf.setQueryFilter(new Equal(
				new BasicFieldOperand(c2),
				new BasicParameterOperand(PTypeSpace.STRING, "?1")));
		leaf.setRdbAdapter(adapter);
		return leaf;
	}

	public QueryLeaf getQueryLeafOrderByA(Example e) throws MedorException {
		RdbClassMapping cm = (RdbClassMapping)
				e.clazz.getClassProject(getProjectName()).getMapping(getMapperName()).getClassMapping();
		String name = e.clazz.getFQName();
		String prefix = name + ".";

		QualifiedTable qt = new BasicQualifiedTable(cm.getRdbTable().getName(), "T_REF_SING_A_0");
		BasicRdbExpQueryLeaf leaf = new BasicRdbExpQueryLeaf(
				null, new QualifiedTable[]{qt}, "");
		leaf.addRdbField(prefix + "id", PTypeSpace.LONG, "id_col", qt);
		RdbExpField f1 = leaf.addRdbField(prefix + "field1", PTypeSpace.STRING, "FIELD_1", qt);
		leaf.removeRdbField(f1);
		RdbExpField f2 = leaf.addRdbField(prefix + "field2", PTypeSpace.STRING, "FIELD_2", qt);
		leaf.removeRdbField(f2);
		OrderField of1 = new BasicOrderField(f1, true);
		OrderField of2 = new BasicOrderField(f2);
		leaf.setOrderBy(new OrderField[] {of1, of2});
		leaf.setQueryFilter(new Equal(
				new BasicFieldOperand(f1),
				new BasicParameterOperand(PTypeSpace.STRING, "?1")));
		leaf.setRdbAdapter(adapter);
		return leaf;
	}

	public QueryLeaf getQueryLeafMemberOfBD(Example eB, Example eD) throws Exception {
		RdbClassMapping cmB = (RdbClassMapping)
				eB.clazz.getClassProject(getProjectName()).getMapping(getMapperName()).getClassMapping();
		RdbClassMapping cmD = (RdbClassMapping)
				eD.clazz.getClassProject(getProjectName()).getMapping(getMapperName()).getClassMapping();
		String name = eB.clazz.getFQName();
		String prefix = name + ".";

		QualifiedTable qtB = new BasicQualifiedTable(cmB.getRdbTable().getName(), "T_B_0");
		BasicRdbExpQueryLeaf leafB = new BasicRdbExpQueryLeaf(
				null, new QualifiedTable[]{qtB}, "");
		RdbExpField fc2 = leafB.addRdbField(prefix + "c2", PTypeSpace.CHAR, "C2", qtB);
		RdbExpField fc1 = leafB.addRdbField(prefix + "c1", PTypeSpace.BYTE, "C1", qtB);

		QualifiedTable qtD = new BasicQualifiedTable(cmD.getRdbTable().getName(), "T_D_1");
		BasicRdbExpQueryLeaf leafD = new BasicRdbExpQueryLeaf(
				null, new QualifiedTable[]{qtD}, "");
		leafD.setRdbAdapter(adapter);
		RdbExpField fb = leafD.addRdbField(prefix + "b", PTypeSpace.BYTE, "B", qtD);
		RdbExpField fa = leafD.addRdbField(prefix + "a", PTypeSpace.CHAR, "A", qtD);

		ArrayList left = new ArrayList(2);
		ArrayList right = new ArrayList(2);
		left.add(new BasicFieldOperand(fc2));
		left.add(new BasicFieldOperand(fc1));
		right.add(new BasicFieldOperand(fb));
		right.add(new BasicFieldOperand(fa));
		MemberOf mo = new MemberOf(left, right);
		leafB.setQueryFilter(mo);
		leafB.setRdbAdapter(adapter);
		return leafB;
	}

	public QueryLeaf getQueryLeafMemberOfBC(Example eB, Example eC) throws Exception {
		RdbClassMapping cmB = (RdbClassMapping)
				eB.clazz.getClassProject(getProjectName()).getMapping(getMapperName()).getClassMapping();
		RdbClassMapping cmC = (RdbClassMapping)
				eC.clazz.getClassProject(getProjectName()).getMapping(getMapperName()).getClassMapping();
		String name = eB.clazz.getFQName();
		String prefix = name + ".";

		QualifiedTable qtB = new BasicQualifiedTable(cmB.getRdbTable().getName(), "T_B_0");
		BasicRdbExpQueryLeaf leafB = new BasicRdbExpQueryLeaf(
				null, new QualifiedTable[]{qtB}, "");
		leafB.setRdbAdapter(adapter);
		RdbExpField fc2 = leafB.addRdbField(prefix + "c2", PTypeSpace.CHAR, "C2", qtB);
		RdbExpField fc1 = leafB.addRdbField(prefix + "c1", PTypeSpace.BYTE, "C1", qtB);
		RdbExpField fy = leafB.addRdbField(prefix + "y", PTypeSpace.LONG, "Y", qtB);
		leafB.removeRdbField(fy);

		QualifiedTable qtC = new BasicQualifiedTable(cmC.getRdbTable().getName(), "T_C_1");
		BasicRdbExpQueryLeaf leafC = new BasicRdbExpQueryLeaf(
				null, new QualifiedTable[]{qtC}, "");
		leafC.setRdbAdapter(adapter);
		RdbExpField fx = leafC.addRdbField(prefix + "x", PTypeSpace.LONG, "X", qtC);
		leafC.setQueryFilter(new Equal(new BasicFieldOperand(fx), new BasicOperand((long) 2)));
		ArrayList left = new ArrayList(1);
		ArrayList right = new ArrayList(1);
		left.add(new BasicFieldOperand(fy));
		right.add(new BasicFieldOperand(fx));
		MemberOf mo = new MemberOf(left, right);
		leafB.setQueryFilter(mo);
		return leafB;
	}

	public QueryLeaf getQueryLeafMOfParamBD(Example eB, Example eD) throws Exception {
		RdbClassMapping cmB = (RdbClassMapping)
				eB.clazz.getClassProject(getProjectName()).getMapping(getMapperName()).getClassMapping();
		RdbClassMapping cmD = (RdbClassMapping)
				eD.clazz.getClassProject(getProjectName()).getMapping(getMapperName()).getClassMapping();
		String nameB = eB.clazz.getFQName();
		String prefixB = nameB + ".";
		String nameD = eD.clazz.getFQName();
		String prefixD = nameD + ".";

		QualifiedTable qtB = new BasicQualifiedTable(cmB.getRdbTable().getName(), "T_B_0");
		QualifiedTable qtD = new BasicQualifiedTable(cmD.getRdbTable().getName(), "T_D_1");
		BasicRdbExpQueryLeaf leafB = new BasicRdbExpQueryLeaf(
				null, new QualifiedTable[]{qtB}, "");
		leafB.setRdbAdapter(adapter);
		BasicRdbExpQueryLeaf leafD = new BasicRdbExpQueryLeaf(
				null, new QualifiedTable[]{qtD}, "");
		leafD.setRdbAdapter(adapter);
		RdbExpField fy = leafB.addRdbField(prefixB + "y", PTypeSpace.LONG, "Y", qtB);
		leafB.removeRdbField(fy);

		RdbExpField fb = leafD.addRdbField(prefixD + "b", PTypeSpace.CHAR, "B", qtD);
		RdbExpField fa = leafD.addRdbField(prefixD + "a", PTypeSpace.BYTE, "A", qtD);

		ArrayList left = new ArrayList(1);
		BasicOperand value = new BasicOperand(1013);
		left.add(value);
		ArrayList right = new ArrayList(1);
		right.add(new BasicFieldOperand(fy));
		MemberOf mo = new MemberOf(left, right);
		leafD.setQueryFilter(mo);
		return leafD;
	}

	public QueryLeaf getQueryLeafAnimal(Example e) throws MedorException {
		RdbClassMapping cm = (RdbClassMapping)
				e.clazz.getClassProject(getProjectName()).getMapping(getMapperName()).getClassMapping();
		String name = e.clazz.getFQName();
		String prefix = name + ".";

		QualifiedTable qt = new BasicQualifiedTable(cm.getRdbTable().getName(), "Animal_0");
		BasicRdbExpQueryLeaf leaf = new BasicRdbExpQueryLeaf(
				null, new QualifiedTable[]{qt}, "");
		leaf.setRdbAdapter(adapter);
		leaf.addRdbField(prefix + "animal_id", PTypeSpace.LONG, "a_id", qt);
		RdbExpField f1 = leaf.addRdbField(prefix + "species", PTypeSpace.STRING, "speciesCol", qt);
		RdbExpField f2 = leaf.addRdbField(prefix + "type", PTypeSpace.STRING, "typeCol", qt);
		leaf.removeRdbField(f1);
		leaf.removeRdbField(f2);
		leaf.setQueryFilter(
				new And(
					new Equal(
						new BasicFieldOperand(f1),
						new BasicOperand("marsupial")),
					new Or(
						new Equal(
							new BasicFieldOperand(f2),
							new BasicOperand("koala")),
						new Equal(
							new BasicFieldOperand(f2),
							new BasicOperand("kangaroo"))
					)
				)
		);
		return leaf;
	}


}
