/**
 * MEDOR: Middleware Enabling Distributed Object Requests
 *
 * Copyright (C) 2001-2004 France Telecom R&D
 * Contact: alexandre.lefebvre@rd.francetelecom.com
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Initial developers: M. Alia, S. Chassande-Barrioz, A. Lefebvre
 */
package org.objectweb.medor.query.jorm;

import org.objectweb.jorm.metainfo.api.GenClassRef;
import org.objectweb.medor.api.Field;
import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.expression.lib.Equal;
import org.objectweb.medor.expression.lib.Or;
import org.objectweb.medor.expression.lib.BasicVariableOperand;
import org.objectweb.medor.filter.lib.BasicFieldOperand;
import org.objectweb.medor.filter.jorm.lib.NavigatorOperator;
import org.objectweb.medor.jorm.TestJormHelper;
import org.objectweb.medor.optim.api.QueryRewriter;
import org.objectweb.medor.optim.jorm.JormFlatten2Rdb;
import org.objectweb.medor.optim.lib.BasicQueryRewriter;
import org.objectweb.medor.optim.lib.FlattenQueryTreeRule;
import org.objectweb.medor.query.TestQueryTreeHelper;
import org.objectweb.medor.query.api.PropagatedField;
import org.objectweb.medor.query.api.QueryNode;
import org.objectweb.medor.query.api.QueryTree;
import org.objectweb.medor.query.api.QueryTreeField;
import org.objectweb.medor.query.jorm.api.JormExtent;
import org.objectweb.medor.query.jorm.lib.ClassExtent;
import org.objectweb.medor.query.jorm.lib.GenClassExtent;
import org.objectweb.medor.query.jorm.lib.JormQueryTreeHelper;
import org.objectweb.medor.query.jorm.lib.NavigatorNodeFactory;
import org.objectweb.medor.query.jorm.lib.PNameField;
import org.objectweb.medor.query.lib.JoinProject;
import org.objectweb.medor.query.lib.MemberOfHelper;
import org.objectweb.medor.query.lib.QueryTreePrinter;
import org.objectweb.util.monolog.api.BasicLevel;

import java.util.ArrayList;
import java.util.Iterator;

/**
 * @author S.Chassande-Barrioz
 */
public abstract class TestNavigatorNodeFactory extends TestJormPDHelper {

    public TestNavigatorNodeFactory(String testName, String loggerName) {
        super(testName, loggerName);
    }

    public NavigatorOperator getNavigatorOperator(Field f, String[] path)
            throws MedorException {
        if (path.length == 0)
            return null;
        NavigatorOperator nav = new NavigatorOperator(new BasicFieldOperand(f),
                path[0]);
        for (int i = 1; i < path.length; i++) {
            nav = new NavigatorOperator(nav, path[i]);
        }
        return nav;
    }

    /**
     * Test the NavigatorNodeFactory.navigate method. - start with a pname of
     * the class - d.f1 where f1 is a primitive field. ==> The result is the
     * same QueryTree
     */
    public void test1a() {
        TestJormHelper.Example e = getExtentOfD();
        QueryTree result = null;
        try {
            Field f = e.extent.getField(getFieldName(e.extent.getName(),
                    e.pnameFieldName));
            String[] path = new String[] { e.clazz.getFQName() + ".f1" };
            NavigatorOperator nav = getNavigatorOperator(f, path);
            result = NavigatorNodeFactory.navigate(nav, null,
                    NavigatorNodeFactory.NAVIGATION).getQueryTree();
            logger
                    .log(BasicLevel.DEBUG,
                            "test1a (d.f1): Constructed QueryTree");
            QueryTreePrinter.printQueryTree(result, logger);
            TestQueryTreeHelper.equals("", e.extent, result, logger);
        } catch (MedorException me) {
            me.printStackTrace();
            fail(me.getMessage());
        }
    }

    /**
     * Test the NavigatorNodeFactory.navigate method. - start with a pname of
     * the class - d.e where e is a reference field. ==> The result is the same
     * QueryTree.
     */
    public void test1b() {
        TestJormHelper.Example e = getExtentOfD();
        QueryTree result = null;
        try {
            Field f = e.extent.getField(getFieldName(e.extent.getName(),
                    e.pnameFieldName));
            String[] path = new String[] { e.clazz.getFQName() + ".e" };
            NavigatorOperator nav = getNavigatorOperator(f, path);
            result = NavigatorNodeFactory.navigate(nav, null,
                    NavigatorNodeFactory.NAVIGATION).getQueryTree();
            logger.log(BasicLevel.DEBUG, "test1b (d.e): Constructed QueryTree");
            QueryTreePrinter.printQueryTree(result, logger);
            TestQueryTreeHelper.equals("", e.extent, result, logger);
        } catch (MedorException me) {
            me.printStackTrace();
            fail(me.getMessage());
        }
    }

    /**
     * Test the NavigatorNodeFactory.navigate method. - start with a pname of
     * the class - d.es where es is a GenClass of E. The result is a join with
     * the GC node.
     */
    public void test1c() {
        TestJormHelper.Example e = getExtentOfD();
        QueryTree result = null;
        try {
            NavigatorNodeFactory.resetNameIndexes();
            Field f = e.extent.getField(getFieldName(e.extent.getName(),
                    e.pnameFieldName));
            String[] path = new String[] { e.clazz.getFQName() + ".setOfE" };
            NavigatorOperator nav = getNavigatorOperator(f, path);
            result = NavigatorNodeFactory.navigate(nav, null,
                    NavigatorNodeFactory.NAVIGATION_INTO_LAST).getQueryTree();
            logger
                    .log(BasicLevel.DEBUG,
                            "test1c (d.es): Constructed QueryTree");
            QueryTreePrinter.printQueryTree(result, logger);

            NavigatorNodeFactory.resetNameIndexes();
            PNameField pnf = (PNameField) e.extent.getTupleStructure()
                    .getField(e.clazz.getFQName() + ".setOfE");
            GenClassRef gcr = (GenClassRef) pnf.getReference();
            GenClassExtent extent = new GenClassExtent(gcr, e.clazz.getFQName()
                    + ".setOfE", "object1", "element");

            JoinProject join = new JoinProject("");

            //'pnf' references the PNameField of the GenClass
            pnf = (PNameField) extent.getField(extent.getPNameFieldName());
            f = e.extent.getField(e.clazz.getFQName() + ".setOfE");
            //add the join condition: pnf should be equal to the
            //PName of the extent
            join.setQueryFilter(new Equal(new BasicFieldOperand(f),
                    new BasicFieldOperand(pnf)));

            //project in 'join' all fields of the inherited node
            Field[] fToAdd = e.extent.getFields();
            for (int i = 0; i < fToAdd.length; i++) {
                join.addPropagatedField(fToAdd[i].getName(), fToAdd[i]
                        .getType(),
                        new QueryTreeField[] { (QueryTreeField) fToAdd[i] });
            }

            //project in 'join' the element field of the GenClassExtent
            Field elemToAdd = extent.getField(extent.getElementFieldName());
            join.addPropagatedField(elemToAdd.getName(), elemToAdd.getType(),
                    new QueryTreeField[] { (QueryTreeField) elemToAdd });

            join.setQueryFilter(new Equal(new BasicFieldOperand(f),
                    new BasicFieldOperand(pnf)));
            logger.log(BasicLevel.DEBUG, "test1c (d.es): Expected QueryTree");
            QueryTreePrinter.printQueryTree(join, logger);

            TestQueryTreeHelper.equals("", join, result, logger);
        } catch (MedorException me) {
            me.printStackTrace();
            fail(me.getMessage());
        }
    }

    /**
     * Test the NavigatorNodeFactory.navigate/IS_EMPTY method. - start with a
     * pname of the class - d.es where es is a GenClass of E. - add isEmpty to
     * the path traversal The result is a join with the GC node.
     */
    public void test1cEmpty() {
        TestJormHelper.Example e = getExtentOfD();
        QueryTree result1, result = null;
        try {
            Field f = e.extent.getField(getFieldName(e.extent.getName(),
                    e.pnameFieldName));
            String[] path = new String[] { e.clazz.getFQName() + ".setOfE" };
            NavigatorOperator nav = getNavigatorOperator(f, path);
            result1 = NavigatorNodeFactory.navigate(nav, null,
                    NavigatorNodeFactory.IS_EMPTY).getQueryTree();
            logger.log(BasicLevel.DEBUG,
                    "test1cEmpty (d.es): Constructed QueryTree");
            QueryTreePrinter.printQueryTree(result1, logger);

            ArrayList rules = new ArrayList(2);
            rules.add(new FlattenQueryTreeRule());
            rules.add(new JormFlatten2Rdb());
            QueryRewriter jqr = new BasicQueryRewriter(rules);

            //assign mapper to all extents
            Iterator extents = JormQueryTreeHelper.getJormExtents(result1)
                    .iterator();
            while (extents.hasNext()) {
                JormExtent ext = (JormExtent) extents.next();
                logger.log(BasicLevel.DEBUG, "Setting mapper for extent "
                        + ext.getName());
                ext.setPMapper(mapper, PROJECT_NAME);
                ext.setProjectName("test");
            }
            result = jqr.transform(result1);
            logger.log(BasicLevel.DEBUG,
                    "test1cEmpty (d.es): Rewritten QueryTree");
            QueryTreePrinter.printQueryTree(result, logger);
            //TODO add expected query result

            /*
             * PNameField pnf = (PNameField)
             * e.extent.getTupleStructure().getField(e.clazz.getFQName() +
             * ".setOfE"); GenClassRef gcr = (GenClassRef) pnf.getReference();
             * GenClassExtent extent = new GenClassExtent(gcr,
             * e.clazz.getFQName() + ".setOfE", "object",
             * NavigatorNodeFactory.GENCLASS_ELEMENT_NAME);
             * 
             * JoinProject join = new JoinProject("", null);
             * 
             * //'pnf' references the PNameField of the GenClass pnf =
             * (PNameField) extent.getField(extent.getPNameFieldName()); f =
             * e.extent.getField(e.clazz.getFQName() + ".setOfE"); //add the
             * join condition: pnf should be equal to the //PName of the extent
             * join.setQueryFilter( new Equal( new BasicFieldOperand(f), new
             * BasicFieldOperand(pnf) ) );
             * 
             * //project in 'join' all fields of the inherited node Field[]
             * fToAdd = e.extent.getFields(); for (int i = 0; i < fToAdd.length;
             * i++) { join.addPropagatedField( fToAdd[i].getName(),
             * fToAdd[i].getType(), new QueryTreeField[]{(QueryTreeField)
             * fToAdd[i]}); }
             * 
             * //project in 'join' the element field of the GenClassExtent Field
             * elemToAdd = extent.getField(extent.getElementFieldName());
             * join.addPropagatedField(elemToAdd.getName(), elemToAdd.getType(),
             * new QueryTreeField[]{(QueryTreeField) elemToAdd});
             * 
             * join.setQueryFilter( new Equal( new BasicFieldOperand(f), new
             * BasicFieldOperand(pnf) ) ); logger.log(BasicLevel.DEBUG,
             * "Expected QueryTree"); QueryTreePrinter.printQueryTree(join,
             * logger);
             * 
             * TestQueryTreeHelper.equals("", join, result, logger);
             */
        } catch (MedorException me) {
            me.printStackTrace();
            fail(me.getMessage());
        }
    }

    /**
     * d.setOfE
     */
    public void test1d() {
        TestJormHelper.Example e = getExtentOfD();
        QueryTree result = null;
        try {
            NavigatorNodeFactory.resetNameIndexes();
            Field f = e.extent.getField(getFieldName(e.extent.getName(),
                    e.pnameFieldName));
            String[] path = new String[] { e.clazz.getFQName() + ".setOfE" };
            NavigatorOperator nav = getNavigatorOperator(f, path);
            result = NavigatorNodeFactory.navigate(nav, "test1d",
                    NavigatorNodeFactory.NAVIGATION_INTO_LAST).getQueryTree();

            logger.log(BasicLevel.DEBUG,
                    "test1d (d.setOfE into last): Constructed QueryTree");
            QueryTreePrinter.printQueryTree(result, logger);

            NavigatorNodeFactory.resetNameIndexes();
            PNameField pnf = (PNameField) e.extent.getTupleStructure()
                    .getField(e.clazz.getFQName() + ".setOfE");
            GenClassRef gcr = (GenClassRef) pnf.getReference();
            GenClassExtent extent = new GenClassExtent(gcr, e.clazz.getFQName()
                    + ".setOfE", "object1", "element");

            JoinProject join = new JoinProject("");

            //'pnf' references the PNameField of the GenClass
            pnf = (PNameField) extent.getField(extent.getPNameFieldName());
            f = e.extent.getField(e.clazz.getFQName() + ".setOfE");
            //add the join condition: pnf should be equal to the
            //PName of the extent
            join.setQueryFilter(new Equal(new BasicFieldOperand(f),
                    new BasicFieldOperand(pnf)));

            //project in 'join' all fields of the inherited node
            Field[] fToAdd = e.extent.getFields();
            for (int i = 0; i < fToAdd.length; i++) {
                join.addPropagatedField(fToAdd[i].getName(), fToAdd[i]
                        .getType(),
                        new QueryTreeField[] { (QueryTreeField) fToAdd[i] });
            }

            //project in 'join' all fields of the GenClassExtent
            // and find the QueryTreeField which represents the element
            Field elemToAdd = extent.getField(extent.getElementFieldName());
            join.addPropagatedField("test1d", elemToAdd.getType(),
                    new QueryTreeField[] { (QueryTreeField) elemToAdd });

            join.setQueryFilter(new Equal(new BasicFieldOperand(f),
                    new BasicFieldOperand(pnf)));
            logger.log(BasicLevel.DEBUG,
                    "test1d (d.setOfE into last): Expected QueryTree");
            QueryTreePrinter.printQueryTree(join, logger);

            TestQueryTreeHelper.equals("", join, result, logger);
        } catch (MedorException me) {
            me.printStackTrace();
            fail(me.getMessage());
        }
    }

    /**
     * Test the NavigatorNodeFactory.navigate method. - start with a field of
     * the class - d.f1 where f1 is a primitive field. ==> The result is a join
     * with the D node
     */
    public void test2a() {
        TestJormHelper.Example e = getExtentOfC();
        QueryTree result = null;
        NavigatorNodeFactory.resetNameIndexes();

        try {
            QueryTreeField f = (QueryTreeField) e.extent.getField(e.clazz
                    .getFQName()
                    + ".d");
            String[] path = new String[] { "f1" };
            NavigatorOperator nav = getNavigatorOperator(f, path);
            result = NavigatorNodeFactory.navigate(nav, null,
                    NavigatorNodeFactory.NAVIGATION).getQueryTree();
            logger.log(BasicLevel.DEBUG,
                    "test2a (c.d.f1): Constructed QueryTree");
            QueryTreePrinter.printQueryTree(result, logger);
            logger.log(BasicLevel.DEBUG, "test2a (c.d.f1): Expected QueryTree");
            QueryTreePrinter.printQueryTree(get2abExpected(e, "f1"), logger);
            TestQueryTreeHelper.equals("", get2abExpected(e, "f1"), result,
                    logger);
        } catch (MedorException me) {
            me.printStackTrace();
            fail(me.getMessage());
        }
    }

    /**
     * Test the NavigatorNodeFactory.navigate method. - start with a field of
     * the class - d.e where e is a reference field. ==> The result is a join
     * with the D node
     */
    public void test2b() {
        TestJormHelper.Example e = getExtentOfC();
        QueryTree result = null;
        NavigatorNodeFactory.resetNameIndexes();

        try {
            Field f = e.extent.getField(e.clazz.getFQName() + ".d");
            String[] path = new String[] { "e" };
            NavigatorOperator nav = getNavigatorOperator(f, path);
            result = NavigatorNodeFactory.navigate(nav, null,
                    NavigatorNodeFactory.NAVIGATION).getQueryTree();
            logger.log(BasicLevel.DEBUG,
                    "test2b (c.d.e): Constructed QueryTree");
            QueryTreePrinter.printQueryTree(result, logger);
            TestQueryTreeHelper.equals("", get2abExpected(e, "e"), result,
                    logger);
        } catch (MedorException me) {
            me.printStackTrace();
            fail(me.getMessage());
        }
    }

    public QueryTree get2abExpected(Example e, String fn) throws MedorException {
        Example ed = getExtentOfD(new String[] { fn });
        ClassExtent extent = ed.extent;
        JoinProject join = new JoinProject("");

        //project in 'join' all fields of the inherited node
        Field[] fToAdd = e.extent.getFields();
        for (int i = 0; i < fToAdd.length; i++) {
            join.addPropagatedField(fToAdd[i].getName(), fToAdd[i].getType(),
                    new QueryTreeField[] { (QueryTreeField) fToAdd[i] });
        }

        fToAdd = extent.getFields();
        for (int i = 0; i < fToAdd.length; i++) {
            join.addPropagatedField(fToAdd[i].getName(), fToAdd[i].getType(),
                    new QueryTreeField[] { (QueryTreeField) fToAdd[i] });
        }

        join.setQueryFilter(new Equal(new BasicFieldOperand(e.extent
                .getField(e.clazz.getFQName() + ".d")), new BasicFieldOperand(
                extent.getField(ed.clazz.getFQName() + ".object1"))));
        return join;
    }

    /**
     * Test the NavigatorNodeFactory.navigate method. - start with a field of
     * the class - d.es where es is a GenClass of E. The result is a
     * join(join(qt, node(D), node(GC)).
     */
    public void test2c() {
        TestJormHelper.Example e = getExtentOfC();
        QueryTree result = null;
        try {
            Field f = e.extent.getField(e.clazz.getFQName() + ".d");
            String[] path = new String[] { "setOfE" };
            NavigatorOperator nav = getNavigatorOperator(f, path);
            result = NavigatorNodeFactory.navigate(nav, null,
                    NavigatorNodeFactory.NAVIGATION_INTO_LAST).getQueryTree();
            logger.log(BasicLevel.DEBUG,
                    "test2c (c.d.setOfE into last): Constructed QueryTree");
            QueryTreePrinter.printQueryTree(result, logger);
        } catch (MedorException me) {
            me.printStackTrace();
            fail(me.getMessage());
        }
    }

    //A, b.c.d.f1
    public void test3a() {
        TestJormHelper.Example e = getExtentOfA();
        QueryTree result = null;
        try {
            Field f = e.extent.getField(getFieldName(e.extent.getName(),
                    e.pnameFieldName));
            String[] path = new String[] { e.clazz.getFQName() + ".b", "c",
                    "d", "f1" };
            NavigatorOperator nav = getNavigatorOperator(f, path);
            logger.log(BasicLevel.DEBUG, "Initial field operand is "
                    + nav.getFieldOperand().getField().getName());
            logger.log(BasicLevel.DEBUG, "Path is " + nav.getPath());
            result = NavigatorNodeFactory.navigate(nav, null,
                    NavigatorNodeFactory.NAVIGATION).getQueryTree();
            logger.log(BasicLevel.DEBUG,
                    "test3a (a.b.c.d.f1): Constructed QueryTree");
            QueryTreePrinter.printQueryTree(result, logger);
        } catch (MedorException me) {
            me.printStackTrace();
            fail(me.getMessage());
        }
    }

    //A, b.c.d.e
    public void test3b() {
        TestJormHelper.Example e = getExtentOfA();
        QueryTree result = null;
        try {
            Field f = e.extent.getField(getFieldName(e.extent.getName(),
                    e.pnameFieldName));
            String[] path = new String[] { e.clazz.getFQName() + ".b", "c",
                    "d", "e" };
            NavigatorOperator nav = getNavigatorOperator(f, path);
            result = NavigatorNodeFactory.navigate(nav, null,
                    NavigatorNodeFactory.NAVIGATION).getQueryTree();
            logger.log(BasicLevel.DEBUG,
                    "test3b (a.b.c.d.e): Constructed QueryTree");
            QueryTreePrinter.printQueryTree(result, logger);
        } catch (MedorException me) {
            me.printStackTrace();
            fail(me.getMessage());
        }
    }

    //A, [A.b, c, d, es]
    public void test3c() {
        TestJormHelper.Example e = getExtentOfA();
        QueryTree result = null;
        try {
            Field f = e.extent.getField(getFieldName(e.extent.getName(),
                    e.pnameFieldName));
            String[] path = new String[] { e.clazz.getFQName() + ".b", "c",
                    "d", "setOfE" };
            NavigatorOperator nav = getNavigatorOperator(f, path);
            result = NavigatorNodeFactory.navigate(nav, null,
                    NavigatorNodeFactory.NAVIGATION_INTO_LAST).getQueryTree();
            logger.log(BasicLevel.DEBUG,
                    "test3c (a.b.c.d.setOfE): Constructed QueryTree");
            QueryTreePrinter.printQueryTree(result, logger);

        } catch (MedorException me) {
            me.printStackTrace();
            fail(me.getMessage());
        }
    }

    //A, [A.b, c, d, es] ISEMPTY + OR
    public void test3cEmpty() {
        TestJormHelper.Example e = getExtentOfA();
        QueryTree result, result1 = null;
        try {
            Field f = e.extent.getField(getFieldName(e.extent.getName(),
                    e.pnameFieldName));
            String[] path = new String[] { e.clazz.getFQName() + ".b", "c",
                    "d", "setOfE" };
            NavigatorOperator nav = getNavigatorOperator(f, path);
            result1 = NavigatorNodeFactory.navigate(nav, null,
                    NavigatorNodeFactory.IS_NOT_EMPTY).getQueryTree();

            //constructing filter with adding "OR f1=12"
            QueryNode r1 = (QueryNode) result1;
            Field f1 = r1.getTupleStructure().getField(
                    getFieldName(e.extent.getName(), "f1"));
            Field prevF1 = ((PropagatedField) f1).getPreviousFields()[0];
            r1.setQueryFilter(new Or(r1.getQueryFilter(),
                    new Equal(new BasicFieldOperand(prevF1),
                            new BasicVariableOperand(12))));

            logger
                    .log(BasicLevel.DEBUG,
                            "test3cEmpty (a.b.c.d.setOfE is not empty): Constructed QueryTree");
            QueryTreePrinter.printQueryTree(result1, logger);

            ArrayList rules = new ArrayList(2);
            rules.add(new FlattenQueryTreeRule());
            rules.add(new JormFlatten2Rdb());
            QueryRewriter jqr = new BasicQueryRewriter(rules);

            //assign mapper to all extents
            Iterator extents = JormQueryTreeHelper.getJormExtents(result1)
                    .iterator();
            while (extents.hasNext()) {
                JormExtent ext = (JormExtent) extents.next();
                logger.log(BasicLevel.DEBUG, "Setting mapper for extent "
                        + ext.getName());
                ext.setPMapper(mapper, PROJECT_NAME);
                ext.setProjectName("test");
            }
            result = jqr.transform(result1);
            logger.log(BasicLevel.DEBUG, "test3cEmpty: Rewritten QueryTree");
            QueryTreePrinter.printQueryTree(result, logger);
            //TODO add expected query result

        } catch (MedorException me) {
            me.printStackTrace();
            fail(me.getMessage());
        }
    }

    public void test3cMemberOf() {
        TestJormHelper.Example e = getExtentOfA();
        QueryTree input = null;
        try {
            Field f = e.extent.getField(getFieldName(e.extent.getName(),
                    e.pnameFieldName));
            String[] path = new String[] { e.clazz.getFQName() + ".b", "c",
                    "d", "setOfE" };
            NavigatorOperator nav = getNavigatorOperator(f, path);
            input = NavigatorNodeFactory.navigate(nav, null,
                    NavigatorNodeFactory.NAVIGATION_INTO_LAST).getQueryTree();
            //TODO add MemberOf
            logger
                    .log(BasicLevel.DEBUG,
                            "test3cMemberOf (a.b.c.d.setOfE into last): Constructed QueryTree");
            QueryTreePrinter.printQueryTree(input, logger);
            ArrayList rules = new ArrayList(2);
            rules.add(new FlattenQueryTreeRule());
            rules.add(new JormFlatten2Rdb());
            QueryRewriter jqr = new BasicQueryRewriter(rules);

            //assign mapper to all extents
            Iterator extents = JormQueryTreeHelper.getJormExtents(input)
                    .iterator();
            while (extents.hasNext()) {
                JormExtent ext = (JormExtent) extents.next();
                logger.log(BasicLevel.DEBUG, "Setting mapper for extent "
                        + ext.getName());
                ext.setPMapper(mapper, PROJECT_NAME);
                ext.setProjectName("test");
            }
            QueryTree result = jqr.transform(input);
            logger.log(BasicLevel.DEBUG, "test3cMemberOf: Rewritten QueryTree");

            QueryTreePrinter.printQueryTree(result, logger);
        } catch (MedorException me) {
            me.printStackTrace();
            fail(me.getMessage());
        }
    }

    //A, [c.d.es]
    public void test4a() {
        TestJormHelper.Example e = getExtentOfA();
        QueryTree result = null;
        try {
            Field f = e.extent.getField(getFieldName(e.extent.getName(), "b"));
            String[] path = new String[] { "c", "d", "f1" };
            NavigatorOperator nav = getNavigatorOperator(f, path);
            result = NavigatorNodeFactory.navigate(nav, null,
                    NavigatorNodeFactory.NAVIGATION).getQueryTree();
            logger.log(BasicLevel.DEBUG,
                    "test4a (a.b.c.d.f1): Constructed QueryTree");
            QueryTreePrinter.printQueryTree(result, logger);
        } catch (MedorException me) {
            me.printStackTrace();
            fail(me.getMessage());
        }
    }

    //A, [c.d.es]
    public void test4b() {
        TestJormHelper.Example e = getExtentOfA();
        QueryTree result = null;
        try {
            Field f = e.extent.getField(getFieldName(e.extent.getName(), "b"));
            String[] path = new String[] { "c", "d", "e" };
            NavigatorOperator nav = getNavigatorOperator(f, path);
            result = NavigatorNodeFactory.navigate(nav, null,
                    NavigatorNodeFactory.NAVIGATION).getQueryTree();
            logger.log(BasicLevel.DEBUG,
                    "test4b (a.b.c.d.e): Constructed QueryTree");
            QueryTreePrinter.printQueryTree(result, logger);
        } catch (MedorException me) {
            me.printStackTrace();
            fail(me.getMessage());
        }
    }

    //A, [c.d.es]
    public void test4c() {
        TestJormHelper.Example e = getExtentOfA();
        QueryTree result = null;
        try {
            Field f = e.extent.getField(getFieldName(e.extent.getName(), "b"));
            String[] path = new String[] { "c", "d", "setOfE" };
            NavigatorOperator nav = getNavigatorOperator(f, path);
            result = NavigatorNodeFactory.navigate(nav, null,
                    NavigatorNodeFactory.NAVIGATION_INTO_LAST).getQueryTree();
            logger.log(BasicLevel.DEBUG,
                    "test4b (a.b.c.d.setOfE into last): Constructed QueryTree");
            QueryTreePrinter.printQueryTree(result, logger);
        } catch (MedorException me) {
            me.printStackTrace();
            fail(me.getMessage());
        }
    }

    //c.d.e and c.d.es
    public void testMemberOf2stepsCDE() throws Exception {
        NavigatorNodeFactory.resetNameIndexes();
        TestJormHelper.Example e = getExtentOfC();
        QueryTree result1, result2 = null;
        try {
            Field f = e.extent.getField(getFieldName(e.extent.getName(),
                    e.pnameFieldName));
            String[] path = new String[] { e.clazz.getFQName() + ".d", "e" };
            NavigatorOperator nav = getNavigatorOperator(f, path);
            result1 = NavigatorNodeFactory.navigate(nav, null,
                    NavigatorNodeFactory.NAVIGATION).getQueryTree();
            logger.log(BasicLevel.DEBUG,
                    "testMemberOf2stepsCDE (c.d.e): Constructed QueryTree");
            QueryTreePrinter.printQueryTree(result1, logger);

            Field fc = result1.getTupleStructure().getField(
                    "org.objectweb.medor.query.jorm.C.object1");
            path = new String[] { "d", "setOfE" };
            NavigatorOperator nav2 = getNavigatorOperator(fc, path);
            result2 = NavigatorNodeFactory.navigate(nav2, null,
                    NavigatorNodeFactory.NAVIGATION_INTO_LAST).getQueryTree();
            logger.log(BasicLevel.DEBUG,
                    "testMemberOf2stepsCDE (c.d.es): Constructed QueryTree");
            QueryTreePrinter.printQueryTree(result2, logger);

            JoinProject jpOnTop = new JoinProject("onTop");
            ArrayList al1 = new ArrayList(1);
            al1.add(new BasicFieldOperand(result2.getTupleStructure().getField(
                    "e")));
            ArrayList al2 = new ArrayList(1);
            al2.add(new BasicFieldOperand(result2.getTupleStructure().getField(
                    "org.objectweb.medor.query.jorm.D.setOfE.element")));

            MemberOfHelper.addMemberOf(jpOnTop, al1, al2);
            jpOnTop
                    .addPropagatedField(
                            "org.objectweb.medor.query.jorm.D.object2",
                            result2.getTupleStructure().getField(
                                    "org.objectweb.medor.query.jorm.D.object2")
                                    .getType(),
                            new QueryTreeField[] { (QueryTreeField) result2
                                    .getTupleStructure()
                                    .getField(
                                            "org.objectweb.medor.query.jorm.D.object2") });
            logger.log(BasicLevel.DEBUG,
                    "testMemberOf2stepsCDE (c.d.es): with MOF on top");
            QueryTreePrinter.printQueryTree(jpOnTop, logger);

        } catch (MedorException me) {
            me.printStackTrace();
            fail(me.getMessage());
        }
    }

    //d.e and d.es
    public void testMemberOf2stepsDE() throws Exception {
        NavigatorNodeFactory.resetNameIndexes();
        TestJormHelper.Example e = getExtentOfD();
        QueryTree result1, result2 = null;
        try {
            Field f = e.extent.getField(getFieldName(e.extent.getName(),
                    e.pnameFieldName));
            String[] path = new String[] { e.clazz.getFQName() + ".e" };
            NavigatorOperator nav = getNavigatorOperator(f, path);
            result1 = NavigatorNodeFactory.navigate(nav, null,
                    NavigatorNodeFactory.NAVIGATION).getQueryTree();
            logger.log(BasicLevel.DEBUG,
                    "testMemberOf2stepsDE (d.e): Constructed QueryTree");
            QueryTreePrinter.printQueryTree(result1, logger);

            Field fc = result1.getTupleStructure().getField(
                    "org.objectweb.medor.query.jorm.D.object");
            path = new String[] { "setOfE" };
            NavigatorOperator nav2 = getNavigatorOperator(fc, path);
            result2 = NavigatorNodeFactory.navigate(nav2, null,
                    NavigatorNodeFactory.NAVIGATION_INTO_LAST).getQueryTree();
            logger.log(BasicLevel.DEBUG,
                    "testMemberOf2stepsDE (d.es): Constructed QueryTree");
            QueryTreePrinter.printQueryTree(result2, logger);

            JoinProject jpOnTop = new JoinProject("onTop");
            ArrayList al1 = new ArrayList(1);
            al1.add(new BasicFieldOperand(result2.getTupleStructure().getField(
                    e.clazz.getFQName() + ".e")));
            ArrayList al2 = new ArrayList(1);
            al2.add(new BasicFieldOperand(result2.getTupleStructure().getField(
                    e.clazz.getFQName() + ".setOfE.element")));

            MemberOfHelper.addMemberOf(jpOnTop, al1, al2);
            jpOnTop.addPropagatedField(e.clazz.getFQName() + ".object", result2
                    .getTupleStructure().getField(
                            e.clazz.getFQName() + ".object").getType(),
                    new QueryTreeField[] { (QueryTreeField) result2
                            .getTupleStructure().getField(
                                    e.clazz.getFQName() + ".object") });
            logger.log(BasicLevel.DEBUG,
                    "testMemberOf2stepsDE (d.es): with MOF on top");
            QueryTreePrinter.printQueryTree(jpOnTop, logger);

        } catch (MedorException me) {
            me.printStackTrace();
            fail(me.getMessage());
        }
    }

}